from __future__ import annotations
from .base_types import *


@dataclass_json
@dataclass
class AddedBy:
    external_urls: ExternalUrls
    id: str
    type: str
    uri: str
    followers: Optional[Followers] = None
    href: Optional[str] = None


@dataclass_json
@dataclass
class ArtistForAlbum:
    external_urls: ExternalUrls
    href: str
    id: str
    name: str
    type: str
    uri: str


@dataclass_json
@dataclass
class LinkedFrom:
    external_urls: ExternalUrls
    href: str
    id: str
    type: str
    uri: str


@dataclass_json
@dataclass
class Owner:
    display_name: Optional[str]
    external_urls: ExternalUrls
    href: str
    id: str
    type: str
    uri: str
    followers: Optional[Followers] = None



from __future__ import annotations
from typing import List
from .base_types import *


@dataclass_json
@dataclass
class CurrentUserProfile:
    display_name: str
    external_urls: ExternalUrls
    href: str
    id: str
    images: List[Image]
    type: str
    uri: str
    followers: Followers
    country: Optional[str] = None
    product: Optional[str] = None
    explicit_content: Optional[ExplicitContent] = None
    email: str = None
    birthdate: Optional[str] = None
    policies: Optional[Policies] = None


@dataclass_json
@dataclass
class UserProfile:
    display_name: str
    external_urls: ExternalUrls
    followers: Followers
    href: str
    id: str
    images: List[Image]
    type: str
    uri: str



import datetime
import json
import re
from typing import Type, TypeVar, Callable, Any

from shrillecho.types.album_types import ArtistAlbums

T = TypeVar('T')


# Respect to spotipy for designing the regex and id / uri parsing functions
_regex_spotify_uri = r'^spotify:(?:(?P<type>track|artist|album|playlist|show|episode):(?P<id>[0-9A-Za-z]+)|user:(?P<username>[0-9A-Za-z]+):playlist:(?P<playlistid>[0-9A-Za-z]+))$'  
_regex_spotify_url = r'^(http[s]?:\/\/)?open.spotify.com\/(?P<type>track|artist|album|playlist|show|episode|user)\/(?P<id>[0-9A-Za-z]+)(\?.*)?$'  
_regex_base62 = r'^[0-9A-Za-z]+$'


def sp_fetch(api_function: Callable, dataclass_type: Type[T], *args, **kwargs) -> T:
    api_response = api_function(*args, **kwargs)
    json_data = json.dumps(api_response)
    return dataclass_type.from_json(json_data)

def get_id(type, id):
    uri_match = re.search(_regex_spotify_uri, id)
    if uri_match is not None:
        uri_match_groups = uri_match.groupdict()
        if uri_match_groups['type'] != type:
            raise Exception( "Unexpected Spotify URI type.")
        return uri_match_groups['id']

    url_match = re.search(_regex_spotify_url, id)
    if url_match is not None:
        url_match_groups = url_match.groupdict()
        if url_match_groups['type'] != type:
            raise Exception( "Unexpected Spotify URL type.")
    
        return url_match_groups['id']

    if re.search(_regex_base62, id) is not None:
        return id
    
    raise Exception( "Unexpected Spotify URL / URI type.")

def is_uri(uri):
    return re.search(_regex_spotify_uri, uri) is not None

def get_uri(type, id):
    if is_uri(id):
        return id
    else:
        return "spotify:" + type + ":" + get_id(type, id)
    
def is_earlier(target_date, min_date):

    try:
        target_date_obj = datetime.strptime(target_date, '%Y-%m-%d')
        min_date_obj = datetime.strptime(min_date, '%Y-%m-%d')
    except:
        return True

    return target_date_obj < min_date_obj


from typing import List, TYPE_CHECKING, Optional
from .base_types import *

@dataclass_json
@dataclass
class Artist:
    external_urls: ExternalUrls
    href: str
    id: str
    name: str
    type: str
    uri: str
    followers: Optional[Followers] = None
    genres: Optional[List[str]] = None
    images: Optional[List[Image]] = None
    popularity: Optional[int] = None

    def func(self):
        print(self.name)


@dataclass_json
@dataclass
class SimpleArtist:
    external_urls: ExternalUrls
    id: str
    name: str
    type: str
    uri: str
    href: Optional[str] = None

@dataclass_json
@dataclass
class FollowerCount:
    followers: int

@dataclass_json
@dataclass
class Cursors:
    after: str
    before: str

@dataclass_json
@dataclass
class FollowedArtistsItems:
    items: List[Artist]


@dataclass_json
@dataclass
class FollowedArtists:
    artists: FollowedArtistsItems


@dataclass_json
@dataclass
class SeveralArtists:
    artists: List[Artist]






import asyncio
import math
import time
from typing import Callable, List, Type

from annotated_types import T


class TaskScheduler:

    def __init__(self):
        self.task_limit = 10
        self.active_task_count = 0
        self.tasks = []

    def load_task(self, task_func: Callable, *args):
        self.tasks.append((task_func, args))

    def load_tasks(self, tasks):
        self.tasks.extend(tasks)

    async def load_task_batch(self):
        current_batch = []
        for task_func, args in self.tasks[:self.task_limit]:
            current_batch.append(task_func(*args))
        return current_batch

    async def execute_tasks(self, unpack=True, batch_size=None):
        if batch_size:
            self.task_limit = batch_size
            
        all_results = []
        tasks_to_rerun = []

        total_tasks = len(self.tasks)
        completed = 0

        while self.tasks:
            # print(f'tasks to complete: {len(self.tasks)}')
            current_batch = []
            for task_func, args in self.tasks[:self.task_limit]:
                current_batch.append(task_func(*args))
        
            results = await asyncio.gather(*current_batch, return_exceptions=Exception)
          
            for (task_func, args), result in zip(self.tasks[:self.task_limit], results):
                if isinstance(result, Exception):
                    print("exception")
                    print(result)
                    tasks_to_rerun.append((task_func, args))
                else:
                    completed+=1
                    print(f"{completed}/{total_tasks}")
                    all_results.append(result)

            self.tasks = self.tasks[self.task_limit:] + tasks_to_rerun
            tasks_to_rerun = []

        if unpack:
            unpacked = []
            for item in all_results:
                if isinstance(item, list):
                    for sub_item in item:
                        unpacked.append(sub_item)
                elif hasattr(item, 'items'):
                    print(item)
                    for sub_item in item.items:
                        if hasattr(sub_item, 'track'):
                            unpacked.append(sub_item.track)
                        else:
                            unpacked.append(sub_item)
            return unpacked
        
        return all_results
        
            


        

    




from typing import Tuple
from io import BytesIO
from itertools import combinations
from typing import List, Set
from dotenv import load_dotenv
from soundcloud import SoundCloud, AlbumPlaylist
from spotipy import Spotify
import requests
from shrillecho.types.soundcloud_types import SoundCloudTrack, UsefulMeta
from shrillecho.types.track_types import Track, TrackSearch
import json
from dataclasses import dataclass
import httpx
import asyncio
from dotenv import load_dotenv
import os
import time
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity
import cv2
from skimage.metrics import structural_similarity as ssim
import numpy as np



class SoundCloudConverter:
    @dataclass
    class ResolvePlaylist:
        name: str
        total_tracks: str
        tracks: List[SoundCloudTrack]

    @dataclass
    class SoundCloudMeta():
        title: str
        description: str
        isrc: str
        artist: str
        username: str
        created_at: str
        duration: int
        full_duration: int
        release_date: str

    __base_url = 'https://api-v2.soundcloud.com'
    __spotify_base_url = 'https://api.spotify.com/v1'

    __soundcloud_headers = {
        "Authorization": "",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/119.0",
    }
   



    def __init__(self, sp: Spotify, sc: SoundCloud, token):
        self.__sp = sp
        self.__sc = sc
        self.__spot_token = token

    @staticmethod
    def get_header():
        with open('.cache', 'r') as file:
            data = json.load(file)
        access_token = data.get('access_token', None)

        headers = {
            "Authorization": f"Bearer {access_token}"
        }

        return headers

    @staticmethod
    def gen_combos(meta: List[str], max_length: int = 3) -> Set[str]:
        
        def clean_tokens(strings, characters_to_remove):
            translation_table = str.maketrans('', '', characters_to_remove)
            cleaned_strings = [s.translate(translation_table) for s in strings]
            stripped_strings = [str.strip() for str in cleaned_strings]
            return stripped_strings
        
        tokens = set(clean_tokens(meta, ",@#()"))
        permuted_combos = set()
        for r in range(1, max_length + 1):
            for c in combinations(tokens, r):
                # for p in permutations(c): // removing permutations as optimisation
                permuted_combos.add(' '.join(c))
        return permuted_combos

    @staticmethod
    def tfidf_similarity(new_string, string_list):
        combined_text = " ".join(string_list)
        documents = [combined_text, new_string]
        tfidf = TfidfVectorizer()
        tfidf_matrix = tfidf.fit_transform(documents)
        similarity = cosine_similarity(tfidf_matrix[0:1], tfidf_matrix[1:2])
        return similarity[0][0]

    async def _get_soundcloud_tracks(self, sc_ids: List[int]) -> List[SoundCloudTrack]:
        async with httpx.AsyncClient() as client:
            responses = await asyncio.gather(
                *(client.get(f"{self.__base_url}/tracks/{sc_id}", headers=self.__soundcloud_headers) for sc_id in
                  sc_ids))
        tracks = [SoundCloudTrack.from_json(json.dumps(resp.json())) for resp in responses]
        return tracks

    async def _resolve_soundcloud_tracks(self, sc_link: str) -> ResolvePlaylist:
        sc_playlist: AlbumPlaylist = self.__sc.resolve(sc_link)
        sc_tracks: List[SoundCloudTrack] = await self._get_soundcloud_tracks(
            [sc_track.id for sc_track in sc_playlist.tracks])
        return SoundCloudConverter.ResolvePlaylist(tracks=sc_tracks, name=sc_playlist.title,
                                                   total_tracks=sc_playlist.track_count)

    async def async_search(self, queries: List[str], search_type: str) -> List[TrackSearch]:
        '''
            Given a set of search combos, for each one return the first 50 results for artist hits.
        '''
        params_list = []
        for q in queries:
            params_list.append(
                {
                    'q': q,
                    'limit': 50,
                    'offset': 0,
                    'type': search_type,
                }
            )

        spot_headers = {
            "Authorization": f"Bearer {self.__spot_token}"
        }

        async with httpx.AsyncClient() as client:
            responses = await asyncio.gather(
                *(client.get(f"{self.__spotify_base_url}/search", headers=spot_headers, params=param) for
                  param in params_list))

        for resp in responses:
            if resp.status_code != 200:
                print(resp.text)
                print("rate limit error")
                exit(0)

        searches = [TrackSearch.from_json(json.dumps(resp.json())) for resp in responses]
        return searches

    async def fetch_pages(self, combos):
        def chunks(lst, n):
            """Yield successive n-sized chunks from lst."""
            for i in range(0, len(lst), n):
                yield lst[i:i + n]

        chunk_size = 50
        sleep = False
        search_combos = combos
        if len(combos) > 100:
            search_combos = list(combos)[20:40]
            chunk_size = 10
            sleep = True
        combo_chunks = [list(chunk) for chunk in chunks(list(search_combos), chunk_size)]
        all_pages: List[TrackSearch] = []
        for combo_chunk in combo_chunks:
            if sleep:
                time.sleep(4)
            all_pages.extend(await self.async_search(combo_chunk, search_type='track'))
        return all_pages

    def analyse_pages(self, all_pages: List[TrackSearch], combos, meta: UsefulMeta):
        ids_added = []
        scores: List[Tuple[int, Track]] = []
        for page in all_pages:
            try:
                for track in page.tracks.items:
                    track_score = SoundCloudConverter.tfidf_similarity(track.name, combos)
                    artist_score = SoundCloudConverter.tfidf_similarity(track.artists[0].name, combos)
                    if track_score > 0.2 and artist_score > 0.1:
                        if track.id not in ids_added:
                            if abs(track.duration_ms - meta.duration) < 1500:
                                scores.append((track_score, track))
                                ids_added.append(track.id)
            except Exception as e:
                print("error", str(e))
                exit(0)

        return scores

    @staticmethod
    def download_image(url):
        response = requests.get(url)
        if response.status_code == 200:
            image_bytes = BytesIO(response.content)
            img = cv2.imdecode(np.frombuffer(image_bytes.read(), np.uint8), -1)
            if img is not None:
                return img
            else:
                raise Exception(f"Failed to decode image from URL: {url}")
        else:
            raise Exception(f"Failed to download image from URL: {url}")

    def compare_images_by_url(self, url1, url2):
        img1 = SoundCloudConverter.download_image(url1)
        img2 = SoundCloudConverter.download_image(url2)
        if img1 is not None and img2 is not None:
            img1_gray = cv2.cvtColor(img1, cv2.COLOR_BGR2GRAY)
            img2_gray = cv2.cvtColor(img2, cv2.COLOR_BGR2GRAY)
            common_size = (100, 100)
            img1_resized = cv2.resize(img1_gray, common_size)
            img2_resized = cv2.resize(img2_gray, common_size)
            similarity_score = ssim(img1_resized, img2_resized)
            return similarity_score

    async def resolve_combinations(self, combos, meta: UsefulMeta) -> Track:

        all_pages: List[TrackSearch] = await self.fetch_pages(combos)
        time.sleep(1)

        scores: List[Tuple[int, Track]] = self.analyse_pages(all_pages, combos, meta)

        if len(scores) > 0:
            for item in scores:
                print(item[1].name, item[0])

            best_score = max(scores, key=lambda x: x[0])

            if best_score[0] < 0.4:
                image_score = self.compare_images_by_url(meta.artwork_url, best_score[1].album.images[2].url)
                if image_score > 0.7:
                    print("passed via image check")
                    return best_score[1]
                else:
                    print("failed image check")
                    return None

            print("passed via being similar enough already")
            return best_score[1]

    async def _search(self, sc_track_meta: UsefulMeta) -> Track:

        combos = SoundCloudConverter.gen_combos(meta=sc_track_meta.meta_list)
        return await self.resolve_combinations(combos, sc_track_meta)

    def _try_isrc_conversion(self, meta: UsefulMeta) -> str | None:
        try:
            isrc_search: TrackSearch = TrackSearch.from_json(json.dumps(self.__sp.search(f'isrc:{meta.isrc}', type='track')))
            return isrc_search.tracks.items[0]
        except:
            return None

    async def _attempt_conversion(self, sc_track_meta, index: int) -> Track:
        sp_track: Track = self._try_isrc_conversion(sc_track_meta)
        if sp_track:
            print(
                f'[ISRC MATCH] SPOTIFY {sp_track.name} | {sp_track.artists[0].name} | SOUNDCLOUD {sc_track_meta.title} | {sc_track_meta.artist}')
            return sp_track
        else:
            print(f'[NO VALID ISRC] {sc_track_meta.title}')
            track: Track = await self._search(sc_track_meta)
            if track:
                print(
                    f'[SEARCH MATCH] SPOTIFY {track.name} | {track.artists[0].name} | SOUNDCLOUD {sc_track_meta.title} | {sc_track_meta.artist}')
                return track
        return None

    async def convert(self, track=None, playlist: str = None):
        if playlist:
            sc_playlist: SoundCloudConverter.ResolvePlaylist = await self._resolve_soundcloud_tracks(playlist)
            converted = 0
            pl_ids = []
            for idx, sc_track in enumerate(sc_playlist.tracks):
                if idx > 50:
                    track: Track = await self._attempt_conversion(sc_track.useful_meta, idx)
                    if track:
                        pl_ids.append(track.id)
                        converted += 1
            self.create_playlist_with_tracks(sc_playlist.name, pl_ids)

        elif track:
            print("[SHRILLECHO] Converting track")
            sc_track = self.__sc.resolve(track)
            tracks: List[SoundCloudTrack] = await self._get_soundcloud_tracks([sc_track.id])
        
            track: Track = await self._attempt_conversion(tracks[0].useful_meta, 0)
            return track

    def create_playlist_with_tracks(self, playlist_name, track_ids):

        playlist = self.__sp.user_playlist_create('alcct3q9gddtkpcyidgx1z9yf', playlist_name, public=True)

        # Split track_ids into batches of 50 (maximum allowed per request)
        batch_size = 50
        for i in range(0, len(track_ids), batch_size):
            batch = track_ids[i:i + batch_size]
            self.__sp.playlist_add_items(playlist['id'], batch)

        print(f'Playlist "{playlist_name}" with {len(track_ids)} tracks created successfully.')




from dataclasses import asdict
import inspect
import json
from typing import Type, Callable, Any, Coroutine


class Cache:

    @staticmethod
    async def cache_response(cache_key: str, 
                             cache_update_function, 
                             expiry: int, cache_store=None, class_type: Type =None):
            
        cached_value = cache_store.get(cache_key)
        if cached_value:
            print(f"getting cached: {cache_key}")

            
            retrieved_value = json.loads(cached_value)
           
            return class_type.from_json(json.dumps(retrieved_value))
        else:
            print(f"refetching {cache_key}")
            new_value = await cache_update_function()

            new_value_dict = asdict(new_value)
        
            cache_store.set(cache_key, json.dumps(new_value_dict), ex=expiry)
      
            return new_value


import re

import spotipy
from pymongo import MongoClient

import shrillecho.utility.archive_maybe_delete.spotify_client as sp_client
import shrillecho.utility.archive_maybe_delete.general as general

from shrillecho.utility.archive_maybe_delete.cache import Cache
from shrillecho.utility.archive_maybe_delete.playlist import Playlist
from shrillecho.types.playlist_types import Playlist as Playlist_t

database = "shrillecho-app"
collection = "cache"

client = MongoClient("mongodb://localhost:27017/")
db = client[database]
collection = db[collection]  



class EveryNoiseSeed:
    __track_uri_regex = r'trackid="([\w]+)"'
    __every_noise_radio_url = 'https://everynoise.com/research.cgi?mode=radio&name=spotify:track'

    def __init__(self, sp: spotipy.Spotify, client: sp_client.SpotifyClient, driver, seed_track: str):
        self.__active_pl = None
        general.log("Setting up every noise seed")
        self.new_isrcs = None
        self.__driver = driver
        self.__seed_track = seed_track
        self.__track_name = ''
        self.__track_artist = ''
        self.__filtered_isrcs = set()
        self.__sp = sp
        self.client = client
        self.__cache = Cache('track_isrcs',  self.__cache_reader, self.__cache_writer)
        self.__fetch_track_data()
        self.__fetched_isrcs = set()

    def __cache_writer(self, cache, *args):
        playlist_uri = args[0]
        pl: Playlist_t = Playlist(self.__sp, Playlist_t(playlist_uri))
        self.__fetched_isrcs = pl.get_all_isrcs()
        for isrc in self.__fetched_isrcs:
            cache.write(f'{isrc[0]},{isrc[1]}\n')

    def __cache_reader(self, cache):
        for isrc in cache:
            try:
                print(isrc)
                values = isrc.split(',')
                self.__fetched_isrcs.add((values[0], values[1]))
            except:
                print("failed")

    def __fetch_track_data(self):
        track_data = self.__sp.track(self.__seed_track)
        self.__track_name = track_data['name']
        self.__track_artist = track_data['artists'][0]['name']

    def __scrape(self):
        url = f'{self.__every_noise_radio_url}:{self.__seed_track}'
        self.__driver.get(url)
        matches = set(re.findall(self.__track_uri_regex, self.__driver.page_source))
        tracks = set()
        for match in matches:
            tracks.add(match) # adds the uris

        return general.convert_uris_to_isrcs(tracks, self.client)

    # Todo make it accept a list of playlists
    def filter_playlists(self, playlists: list[str]):
        for playlist_uri in playlists:

            if self.__cache.fetch(f'{playlist_uri}_isrc.txt', playlist_uri):
                print("cache hit, reading")
            else:
                print("cache miss, writing")

            self.__filtered_isrcs.update(self.__fetched_isrcs)

    def write_playlist(self, scrape_depth=2):

        # Collect 10 iterations of every noise
        scraped_tracks = set()
        for i in range(0, scrape_depth):
            scraped_tracks.update(self.__scrape())

        # Remove the isrcs we do not want to see
        scraped_tracks.difference_update(self.__filtered_isrcs)

        # # Convert the isrc back
        # tracks_u = general.convert_isrcs_to_uris(self.__sp, scraped_isrcs)
        general.write_songs_to_playlist(self.__sp, f'EN seed: {self.__track_name} - {self.__track_artist}')



import base64
import json
import random
import string
import webbrowser
from http.server import BaseHTTPRequestHandler, HTTPServer
import threading
import urllib.parse
import requests


def generate_random_string(length):
    """Generate a random string of specified length."""
    return ''.join(random.choices(string.ascii_letters + string.digits, k=length))

def authenticate_url(sp_auth):
 
    state = generate_random_string(16)

    spotify_authorize_url = 'https://accounts.spotify.com/authorize?' + urllib.parse.urlencode({
        'response_type': 'code',
        'client_id': sp_auth.client_id,
        'scope': sp_auth.scope,
        'redirect_uri': sp_auth.redirect_uri,
        'state': state
    })

    return spotify_authorize_url


sp_auth_obj = None

class RequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
    
        parsed_path = urllib.parse.urlparse(self.path)
        query = urllib.parse.parse_qs(parsed_path.query)
        code = query.get('code', [None])[0]
        state = query.get('state', [None])[0]

        if state is None:
            self.send_response(400)
            self.end_headers()
            self.wfile.write(b"State mismatch error.")
            return

        auth_options = {
            'url': 'https://accounts.spotify.com/api/token',
            'data': {
                'code': code,
                'redirect_uri': sp_auth_obj.redirect_uri,
                'grant_type': 'authorization_code',
            },
            'headers': {
                'Content-Type': 'application/x-www-form-urlencoded',
                'Authorization': 'Basic ' + base64.urlsafe_b64encode(f"{sp_auth_obj.client_id}:{sp_auth_obj.client_secret}".encode()).decode()
            }
        }

        response = requests.post(auth_options['url'], data=auth_options['data'], headers=auth_options['headers'])

        if response.ok:
            tokens = response.json()
            with open('.shrillecho_auth_cache', 'w') as file:
                json.dump(tokens, file)

            self.send_response(200)
            self.end_headers()
            self.wfile.write(b'<h1>Authorization successful. You can close this window.</h1>')

            threading.Thread(target=self.server.shutdown).start()  
        else:
            self.send_response(400)
            self.end_headers()
            self.wfile.write(b"Failed to retrieve the access token from Spotify.")

def run_server():
    with HTTPServer(('localhost', 8081), RequestHandler) as server:
        print('Starting server, use <Ctrl-C> to stop')
        server.serve_forever()

def authenticate_local(sp_auth):
    global sp_auth_obj
    sp_auth_obj = sp_auth
    webbrowser.open(authenticate_url(sp_auth))
    run_server()


from dataclasses import dataclass
from typing import List, Optional

from shrillecho.types.base_types import (
    ExternalUrls, Followers, ExternalIds, Copyright, ExplicitContent, Image, Restrictions
)
from shrillecho.types.track_types import Track


# @dataclass
# class AddedBy:
#     external_urls: ExternalUrls
#     followers: Followers
#     href: str
#     id: str
#     type: str
#     uri: str
#
#
# @dataclass
# class ArtistForAlbum:
#     external_urls: ExternalUrls
#     href: str
#     id: str
#     name: str
#     type: str
#     uri: str
#
#
# @dataclass
# class LinkedFrom:
#     external_urls: ExternalUrls
#     href: str
#     id: str
#     type: str
#     uri: str
#
#
# @dataclass
# class Owner:
#     external_urls: ExternalUrls
#     followers: Followers
#     href: str
#     id: str
#     type: str
#     uri: str
#     display_name: Optional[str] = None


# @dataclass
# class Artist:
#     external_urls: ExternalUrls
#     followers: Followers
#     genres: Optional[List[str]]
#     href: str
#     id: str
#     images: List[Image]
#     name: str
#     popularity: int
#     type: str
#     uri: str


# @dataclass
# class SimplifiedTrack:
#     """ track object without its album basically (at least for now)"""
#     artists: List[Artist]
#     available_markets: List[str]
#     disc_number: int
#     duration_ms: int
#     explicit: bool
#     external_urls: ExternalUrls
#     href: str
#     id: str
#     is_playable: bool
#     linked_from: LinkedFrom
#     restrictions: Restrictions
#     name: str
#     preview_url: str
#     track_number: int
#     type: str
#     uri: str
#     is_local: bool


# @dataclass
# class AlbumTracks:
#     """ https://developer.spotify.com/documentation/web-api/reference/get-an-album tracks object """
#     href: str
#     limit: int
#     next: str
#     offset: int
#     previous: str
#     total: int
#     items: List[SimplifiedTrack]


# @dataclass
# class Album:
#     """" Get album - https://developer.spotify.com/documentation/web-api/reference/get-an-album """
#     album_type: str
#     total_tracks: int
#     available_markets: List[str]
#     external_urls: ExternalUrls
#     href: str
#     id: str
#     images: List[Image]
#     name: str
#     release_date: str
#     release_date_precision: str
#     restrictions: Restrictions
#     type: str
#     uri: str
#     artists: List[Artist]
#     tracks: AlbumTracks
#     copyrights: List[Copyright]
#     external_ids: ExternalIds
#     genres: List[str]
#     label: str
#     popularity: int


# @dataclass
# class SeveralAlbums:
#     """ Get several albums - https://developer.spotify.com/documentation/web-api/reference/get-multiple-albums"""
#     albums: List[Album]


# @dataclass
# class SimpleArtist:
#     external_urls: ExternalUrls
#     href: str
#     id: str
#     name: str
#     type: str
#     uri: str

#
# @dataclass
# class Track:
#     """ Get track - https://developer.spotify.com/documentation/web-api/reference/get-track """
#     album: Album
#     artists: List[Artist]
#     available_markets: List[str]
#     disc_number: int
#     duration_ms: int
#     explicit: bool
#     external_urls: ExternalUrls
#     external_ids: ExternalIds
#     href: str
#     id: str
#     is_playable: bool
#     linked_from: LinkedFrom
#     restrictions: Restrictions
#     name: str
#     preview_url: str
#     track_number: int
#     type: str
#     uri: str
#     is_local: bool


# @dataclass
# class PlaylistTrack:
#     added_at: str
#     added_by: AddedBy
#     is_local: bool
#     track: Track
#
#
# @dataclass
# class PlaylistTracks:
#     href: str
#     limit: int
#     next: str
#     offset: int
#     previous: str
#     total: int
#     items: List[PlaylistTrack]
#
#
# @dataclass
# class Playlist:
#     """ Get playlist - https://developer.spotify.com/documentation/web-api/reference/get-playlist """
#     collaborative: bool
#     description: str
#     external_urls: ExternalUrls
#     followers: Followers
#     href: str
#     id: str
#     images: List[Image]
#     name: str
#     owner: Owner
#     public: bool
#     snapshot_id: str
#     tracks: PlaylistTracks  # Pagination of PlaylistTrack objects
#     type: str
#     uri: str


# @dataclass
# class SeveralTracks:
#     """ Get several tracks - https://developer.spotify.com/documentation/web-api/reference/get-several-tracks"""
#     items: List[Track]
#
#
#
# @dataclass
# class SavedTrack:
#     added_at: str
#     track: Track

#
# @dataclass
# class CurrentUserProfile:
#     country: str
#     display_name: str
#     email: str
#     explicit_content: ExplicitContent
#     external_urls: ExternalUrls
#     followers: Followers
#     href: str
#     id: str
#     images: List[Image]
#     product: str
#     type: str
#     uri: str
#
#
# @dataclass
# class UserProfile:
#     display_name: str
#     external_urls: ExternalUrls
#     followers: "Followers"
#     href: str
#     id: str
#     images: List[Image]
#     type: str
#     uri: str


# @dataclass
# class SavedTracks:
#     """ https://developer.spotify.com/documentation/web-api/reference/get-users-saved-tracks """
#     href: str
#     limit: int
#     next: str
#     offset: int
#     previous: str
#     total: int
#     items: List[SavedTrack]



import asyncio
from dataclasses import asdict, dataclass
import math
import random
import time
from typing import List, Optional, Type, Union
from annotated_types import T
from fastclasses_json import dataclass_json
import httpx
import json
import os

import requests
import strawberry

from shrillecho.auth.local_auth import authenticate_local
from shrillecho.auth.oauth import ClientCredentials, OAuthCredentials
from shrillecho.spotify.task_scheduler import TaskScheduler
from shrillecho.types.album_types import Album, AlbumTracks, ArtistAlbums, SeveralAlbums, SimplifiedAlbum, SimplifiedTrack
from shrillecho.types.artist_types import Artist, FollowedArtists, SeveralArtists
from shrillecho.types.playlist_types import Playlist, PlaylistTrack, PlaylistTracks, SimplifiedPlaylistObject, UserPlaylists
from shrillecho.types.soundcloud_types import User
from shrillecho.types.track_types import AudioFeatures, Recc, SavedTrack, SavedTracks, SeveralTracks, Track
from shrillecho.types.user_types import CurrentUserProfile
from shrillecho.utility.archive_maybe_delete.spotify_client import SpotifyClientGlitch
from shrillecho.utility.general_utility import get_id, sp_fetch
import spotipy
# from shrillecho.auth.local_auth import authenticate_local
from shrillecho.utility.archive_maybe_delete.old_cache import Cache
from shrillecho.utility.cache import cached, mongo_client, redis_client
import redis

@dataclass
class SpotifyAuthContext:
    client_id: str
    client_secret: str
    scope: str
    redirect_uri: str


class SpotifyClient:

    __spotify_api = "https://api.spotify.com/v1/"

    def __init__(self, auth=None, auth_flow: Union[OAuthCredentials | ClientCredentials] = None):
        self.auth = auth
        self.auth_flow = auth_flow
        self.access_token = None
        if auth_flow:
            self.access_token = auth_flow.get_access_token()
        else:
            self.access_token = self.auth
        self.temp_spotipy: spotipy.Spotify = spotipy.Spotify(auth=self.access_token)
        self.http_client: httpx.AsyncClient = httpx.AsyncClient()
        self.redis = redis.Redis(host='localhost', port=6379, db=0)
        self.batch_errors = 0
        self.rate_lock = asyncio.Event()
        glitch = SpotifyClientGlitch()
        self.glitched_token = glitch.access_token
        self.task_scheduler = TaskScheduler()
      

        self.batch_param_map = {}
        self.batch_param_map[SavedTracks] = ()
        # self.glitched_client_token = glitch.client_token
      
    
    class ShrillechoException(Exception):
        def __init__(self, message, error_code=None, retry_after: int = None):
            super().__init__(message)
            self.error_code = error_code
            self.retry_after = retry_after


    def create_get_routine(self, url, resp_type, req_proxy):
        return (self._get, (url, resp_type, req_proxy))
       
    async def _req_proxy(self,
                       method: str, 
                       endpoint: str, 
                       response_type = None,
                       params: Optional[dict] = None, 
                       body: Optional[dict] = None):
        

        data = {
            "params": params,
            "body": body,
            "method":  method,
            "endpoint": f"{self.__spotify_api}{endpoint}"
        }


        response = await self.http_client.post("http://localhost:8002/req", json=data)
        resp_json = response.json()
        data= resp_json['data']
        response.status_code = resp_json['status_code']
        retry_after = resp_json['retry_after']
      
        if response.status_code > 300:
            retry_after = None
            if response.status_code == 429:
                self.rate_lock.set()
                retry_after = int(retry_after)
                print(f"sleeping setting lock: {retry_after} seconds")
                await asyncio.sleep(retry_after)
                self.rate_lock.clear()
            
            print(response.text)
            raise SpotifyClient.ShrillechoException("Spotify Client Error", error_code=response.status_code, 
                                                    retry_after=retry_after)
        if response_type:
            return response_type.from_json(json.dumps(data))
        else:
            return response.json()
               

    async def _request(self,
                       method: str, 
                       endpoint: str, 
                       response_type = None,
                       params: Optional[dict] = None, 
                       body: Optional[dict] = None, req_proxy: bool = True):
        
   
        req_proxy = False
        if req_proxy:
     
      
           return await self._req_proxy(method=method, endpoint=endpoint, response_type=response_type, params=params, body=body)
        
        headers = {"authorization" : f"Bearer {self.access_token}"}
        
        while self.rate_lock.is_set():
            await asyncio.sleep(1)

        print(f"endpoint: {self.__spotify_api}{endpoint}")
        response = await self.http_client.request(method, f"{self.__spotify_api}{endpoint}", params=params, headers=headers, json=body)
        if response.status_code > 300:
            retry_after = None
            if response.status_code == 429:
                self.rate_lock.set()
                print(response.status_code, response.headers.get('Retry-after'))
                try:
                    retry_after = int(response.headers.get('Retry-After'))
                    print(f"sleeping setting lock: {retry_after} seconds")
                except:
                    
                    retry_after = 5
                await asyncio.sleep(retry_after)
                self.rate_lock.clear()
           
            # print(response.text)
            # raise SpotifyClient.ShrillechoException("Spotify Client Error", error_code=response.status_code, 
            #                                         retry_after=retry_after)
        if response_type:
            return response_type.from_json(json.dumps(response.json()))
        else:
            return response.json()
        
    async def _get(self, endpoint: str, response_type=None,  req_proxy: bool = True , params: Optional[dict]=None):
     
        return await self._request(method="GET", endpoint=endpoint, params=params, response_type=response_type, req_proxy=req_proxy)
    
    async def _post(self, endpoint: str, body: dict, params: Optional[dict]=None):
        await self._request(method="POST", endpoint=endpoint, params=params, body=body)

    ############# Shrillecho Implemented Methods #############

    @cached(cache_key_prefix="me", class_type=CurrentUserProfile, expiry=3600)
    async def me(self) -> User:
        return await self._get("me", response_type=CurrentUserProfile, req_proxy=False)
    
    async def track(self, track: str) -> Track:
        return await self._get(f"tracks/{get_id('track', track)}", response_type=Track)

    @cached(cache_key_prefix="album", class_type=Album, expiry=3600)
    async def album(self, album: str) -> Track:
        return await self._get(f"albums/{get_id('album', album)}", response_type=Album)

    @cached(cache_key_prefix="artist", class_type=Artist, expiry=3600)
    async def artist(self, artist: str) -> Track:
        return await self._get(f"artists/{get_id('artist', artist)}", response_type=Artist)

    async def tracks(self, tracks) -> List[Track]:
        tlist = [get_id("track", t) for t in tracks]
        s_tracks: SeveralTracks = await self._get("tracks/?ids=" + ",".join(tlist), response_type=SeveralTracks)
        return s_tracks.tracks

    @cached(cache_key_prefix="artist_related", class_type=Artist, expiry=1000)
    async def artist_related(self, artist: str) -> List[Artist]:
        artist_id = get_id("artist", artist)
        several_artists: SeveralArtists = await self._get(f"artists/{artist_id}/related-artists", response_type=SeveralArtists)
        return several_artists.artists

    async def recccs(self) -> Recc:
        return await self._get("recommendations?limit=50&seed_tracks=5m4Lh4bgiYKSj2MlDcrnxh", response_type=Recc)
    
    async def artists(self, artists: List[str]) -> SeveralArtists:
        artist_ids = [get_id("artist", a) for a in artists]
        return await self._get("artists?ids=" + ",".join(artist_ids), response_type=SeveralArtists)
    
    async def artist_albums_single(self, artist, limit=50, offset=None) -> ArtistAlbums:
        return await self._get(f"artists/{get_id('artist', artist)}/albums?limit={limit}&offset={offset}", response_type=ArtistAlbums)
        
    async def current_user_followed_artists(self, limit: int = 20, after=None) -> FollowedArtists:
       
        return await self._get(f"me/following?type=artist&limit={limit}&after={after}", response_type=FollowedArtists, req_proxy=False)
       
    async def audio_features(self, track) -> AudioFeatures:
        return await self._get(f"audio-features/{get_id('track', track)}", response_type=AudioFeatures, req_proxy=False)

    ################## Spotipy Methods ####################

    async def followed_artists(self):
        self.temp_spotipy.current_user_followed_artists() 
        return sp_fetch(self.temp_spotipy.current_user_following_artists)

    @cached(cache_key_prefix="playlist", class_type=Playlist, expiry=3600)
    async def playlist(self, playlist_id: str) -> Playlist:
        return sp_fetch(self.temp_spotipy.playlist, Playlist, get_id("playlist", playlist_id))
        
    @cached(cache_key_prefix="artist_related", class_type=SeveralArtists, expiry=3600)
    async def artist_related_artists(self, artist: str) -> List[Artist]:
        
        return sp_fetch(self.temp_spotipy.artist_related_artists, SeveralArtists, artist)
       
    # @cached(cache_key_prefix="artist", class_type=Artist, expiry=3600)
    # async def artist(self, artist_id: str) -> Artist:
    #     return sp_fetch(self.temp_spotipy.artist, Artist, get_id("artist", artist_id))
        
    # async def artist_albums_single(self, artist, limit=50, offset=None) -> ArtistAlbums:
    #     return sp_fetch(self.temp_spotipy.artist_albums, ArtistAlbums, artist, limit=limit, offset=offset)
    
    async def albums(self, albums) -> SeveralAlbums:
        return sp_fetch(self.temp_spotipy.albums, SeveralAlbums, albums)
    
    # async def tracks(self, tracks) -> SeveralTracks:
    #     return sp_fetch(self.temp_spotipy.tracks, SeveralTracks, tracks)

    async def user_playlist_create(self, user:str , name:str, public: bool):
        return self.temp_spotipy.user_playlist_create(user=user, name=name, public=public)

    async def playlist_add_items(self, playlist_id: str, tracks) -> None:
        self.temp_spotipy.playlist_add_items(playlist_id, tracks)

    async def current_user_unfollow_playlist(self, playlist_uri) -> None:
        self.temp_spotipy.current_user_unfollow_playlist(playlist_uri)

    async def reccomendations(self, limit, seed_artists=None,seed_genres=None,seed_tracks=None,country=None, track_list=True) -> Recc | List[Track]:

        reccs:  Recc = sp_fetch(self.temp_spotipy.recommendations, 
                              Recc, 
                              limit=limit, 
                              seed_tracks=seed_tracks, 
                              seed_artists=seed_artists, seed_genres=seed_genres)
        if track_list:
            return reccs.tracks
    
        return reccs 

                
    async def current_user_followed_artists(self, limit: int = 20, after=None) -> FollowedArtists :
        return sp_fetch(self.temp_spotipy.current_user_followed_artists, FollowedArtists, limit=limit, after=after)
    
    ################### BATCHING METHODS ############################
        
    @staticmethod
    def generate_pagination_batch_urls(ep_path: str, initial_item , query_params = None) -> List[str]:
        urls = []
        limit = 50
        pages = math.ceil(initial_item.total / limit)
        for page in range(0, pages):
            urls.append(f"{ep_path}?limit=50&offset={limit * page}")
        return urls
    
    async def batch_pagination_tasks(self, class_type, batch_size=5, *args, **kwargs):
        """
            Utilises the task scheduler to batch paginated pages, only works with endpoints
            that have the items, total, next, limit, offset attributes i.e most pagination endpoints

            NOTE: This unpacks by default the results and expands `tracks` attribute to track, consider making this optional to a granular level
        """

        req_proxy = True
        if class_type == SavedTracks:
            ep_path = "me/tracks"
            req_proxy = False
            initial_item = await self._get(f"me/tracks?offset={0}&limit={50}", response_type=SavedTracks, req_proxy=False )

        elif class_type == PlaylistTracks:
            playlist_id = kwargs.get('playlist_id')
            ep_path = f"playlists/{playlist_id}/tracks"
            initial_item = await self._get(f"playlists/{playlist_id}/tracks?offset={0}&limit={50}", response_type=PlaylistTracks)

        elif class_type == UserPlaylists:
            
            if kwargs.get('user_id'):
                user = kwargs.get('user_id')
                ep_path = f"users/{user}/playlists"
                user = kwargs.get('user_id')
                initial_item = await self._get(f"users/{user}/playlists?offset={0}&limit={50}", response_type=UserPlaylists)
            else:
                ep_path = "me/playlists"
               
                req_proxy = False
                initial_item = await self._get(f"me/playlists?offset={0}&limit={50}", response_type=UserPlaylists, req_proxy=False)

        elif class_type == AlbumTracks:
            album_id = kwargs.get('album_id')
            ep_path = f"albums/{album_id}/tracks"
            initial_item = await self._get(f"albums/{album_id}/tracks?offset={0}&limit={50}", response_type=AlbumTracks)

        elif class_type == ArtistAlbums:
            artist = kwargs.get("artist_id")
    
            ep_path = f"artists/{artist}/albums"
            initial_item = await self._get(f"artists/{artist}/albums?offset={0}&limit={50}", response_type=ArtistAlbums)

        urls = SpotifyClient.generate_pagination_batch_urls(ep_path=ep_path, 
                                                                  initial_item=initial_item )
        tasks = [
            self.create_get_routine(url, class_type, req_proxy=req_proxy) for url in urls
        ]

        self.task_scheduler.load_tasks(tasks)

        return await self.task_scheduler.execute_tasks(batch_size=batch_size, unpack=True)

    @cached(cache_key_prefix="saved_tracks_0", class_type=Track)
    async def current_user_saved_tracks(self) -> List[Track]:
        """
            LIKED SONGS
        """
        return await self.batch_pagination_tasks(class_type=SavedTracks, batch_size=5)
    
    @cached(cache_key_prefix="my_playlists", class_type=SimplifiedPlaylistObject)
    async def current_user_saved_playlists(self) -> List[SimplifiedPlaylistObject]:
        """
            CURRENT USER PLAYLISTS 
        """
        return await self.batch_pagination_tasks(class_type=UserPlaylists, batch_size=1)
    
    @cached(cache_key_prefix="user_playlist", class_type=SimplifiedPlaylistObject)
    async def user_playlists(self, user: str):
        """
            SPECIFIC USER PLAYLISTS
        """
        return await self.batch_pagination_tasks(class_type=UserPlaylists, batch_size=5, user_id=user)
    
    @cached(cache_key_prefix="playlist_tracksss", class_type=Track)
    async def playlist_tracks(self, playlist_id: str):
        """
            TRACKS FROM A PLAYLIST

        """
        return await self.batch_pagination_tasks(class_type=PlaylistTracks, batch_size=50, playlist_id=get_id("playlist", playlist_id))
    
    @cached(cache_key_prefix="album_tracksss", class_type=SimplifiedTrack)
    async def album_tracks(self, album: str):
        """
            TRACKS FROM AN ALBUM
        """
    
      
        return await self.batch_pagination_tasks(class_type=AlbumTracks, batch_size=2, album_id=get_id('album', album))
    
    @cached(cache_key_prefix="artist_albumss", class_type=SimplifiedAlbum)
    async def artist_albums(self,artist: str):
        """
            ALBUMS FROM AN ARTIST
        """
    
        return await self.batch_pagination_tasks(class_type=ArtistAlbums, batch_size=2, artist_id=artist)





import spotipy

class SpotifyPlayer:

    @staticmethod
    def play_track(sp: spotipy.Spotify, uri: str):

        """
            Play a specific uri on the first device ID listed.
        """

        uri = f'spotify:track:{uri}'

        device_id = sp.devices()['devices'][0]['id']
        sp.start_playback(device_id=device_id, uris=[uri])


import re

class Parser:

    @staticmethod
    def parse_id_from_url(url: str):
        
        """ Given a spotify url for a playlist, obtain the ID """
        playlist_regex = r'(?:https:\/\/open\.spotify\.com\/playlist\/)([a-zA-Z0-9]+)'
        matches = re.match(playlist_regex, url)
        if matches:
            return matches.group(1)
        else:
            return False
        

    @staticmethod
    def get_uri(track_link: str):
        """ Extract uri from link via regex """
        track_regex = r"https:\/\/open\.spotify\.com\/track\/([a-zA-Z0-9]+)"

        match = re.search(track_regex, track_link)
        if match:
            uri = match.group(1)
            return uri
        else:
            print("No URI found.")



from typing import List
from shrillecho.playlist.virtual_playlist import TracksController
from shrillecho.spotify.client import SpotifyClient
from shrillecho.types.playlist_types import Playlist
from shrillecho.types.track_types import Track


class PlaylistController:

    def __init__(self, sp: SpotifyClient, playlist: Playlist = None, tracks: TracksController = None):

        if not playlist or not tracks:
            raise Exception("Use `create` methods or pass track controller and playlist") 

        self._sp: SpotifyClient = sp
        self.playlist: Playlist = playlist
        self.playlist_tracks: TracksController = tracks
        
    @staticmethod
    async def create(sp: SpotifyClient, playlist_id: str = None, playlist: Playlist = None):

        if not playlist:
            playlist = await sp.playlist(playlist_id=playlist_id)
        
        if not playlist_id:
            playlist_id = playlist.id 

        tracks = await TracksController.create_from_playlist(sp, playlist=playlist_id)

        return PlaylistController(sp=sp, playlist=playlist, tracks=tracks)
    
    def internal_tracks(self) -> List[Track]:
        return self.playlist_tracks.tracks


from typing import List


# Types
from .component_types import AddedBy, Owner
from .base_types import *
from .track_types import Track, TrackInfo
import json

from .user_types import UserProfile


@dataclass_json
@dataclass
class PlaylistTrack:
    added_at: str
    added_by: AddedBy
    is_local: bool
    track: Track
    primary_color: Optional[str]
   


@dataclass_json
@dataclass
class PlaylistTracks:

    limit: Optional[int] = None
    next: Optional[str] = None
    offset: Optional[int]= None
    previous: Optional[str] = None
    total: Optional[int] = None
    items: List[PlaylistTrack] = None
    href: Optional[str] = None


@dataclass_json
@dataclass
class Playlist:
    """ Get playlist - https://developer.spotify.com/documentation/web-api/reference/get-playlist """
    id: str
    tracks: Optional[PlaylistTracks]
    collaborative: Optional[bool] = None
    description: Optional[str] = None
    external_urls: Optional[ExternalUrls] = None
    href: Optional[str] = None
    images: Optional[List[Image]] = None
    name: Optional[str] = None
    owner: Optional[Owner] = None
    primary_color: Optional[str] = None
    public: Optional[bool] = None
    snapshot_id: Optional[str] = None
    type: Optional[str] = None
    uri: Optional[str] = None
    followers: Optional[Followers] = None


@dataclass_json
@dataclass
class SimplifiedPlaylistObject:
    collaborative: bool
    description: str
    external_urls: ExternalUrls
    href: str
    id: str
    images: list[Image]
    name: str
    owner: UserProfile
    public: bool
    snapshot_id: str
    tracks: TrackInfo
    type: str
    uri: str

@dataclass_json
@dataclass
class UserPlaylists:
    """ https://developer.spotify.com/documentation/web-api/reference/get-a-list-of-current-users-playlists """
    href: str
    limit: int
    next: str
    offset: int
    previous: str
    total: int
    items: List[SimplifiedPlaylistObject]



from typing import List

from shrillecho.types.playlist_types import Playlist, PlaylistTracks, PlaylistTrack
import json
import spotipy
import re
from shrillecho.utility.general_utility import *


class PlaylistEngine:

    def __init__(self, sp: spotipy.Spotify):
        self._sp = sp

    def load_playlist_tracks(self, playlist_id: str, load_all_tracks=True) -> List[PlaylistTrack]:
        playlist: Playlist = sp_fetch(self._sp.playlist, Playlist, playlist_id)
        return self._load_tracks(playlist) if load_all_tracks else playlist

    def sort_by_release_date(self, playlist: Playlist):
        track_uris: List[str] = [item.track.uri for item in playlist.tracks.items]
        print(f'number of uris: {len(track_uris)}')
        self._sp.playlist_reorder_items(playlist.id, range_start=0, insert_before=20, range_length=10)
        self._sp.playlist_change_details(playlist.id, description="[shrillecho] - sort test")

    # optimise this do not return a playlist object but rather an array of PlaylistTrack , making the metdata seperate
    def _load_tracks(self, playlist: Playlist) -> List[PlaylistTrack]:
        tracks: List[PlaylistTrack] = []

        playlist_tracks: PlaylistTracks = playlist.tracks

        while playlist_tracks.next:
            limit: int = playlist_tracks.limit
            offset: int = playlist_tracks.offset + limit

            playlist_tracks: PlaylistTracks = sp_fetch(self._sp.playlist_items, PlaylistTracks,
                                                       playlist.id, limit=limit, offset=offset)

            tracks.extend(playlist_tracks.items)

        # add the original fetch tracks
        tracks.extend(playlist.tracks.items)
        return tracks

    def _playlist_tracks_paginate(self, playlist_id, limit, offset) -> PlaylistTracks:
        return sp_fetch(self._sp.playlist_items, PlaylistTracks, playlist_id, limit=limit, offset=offset)

    @staticmethod
    def parse_id_from_url(url: str):
        """ Given a spotify url for a playlist, obtain the ID """
        regex = r'(?:https:\/\/open\.spotify\.com\/playlist\/)([a-zA-Z0-9]+)'
        matches = re.match(regex, url)
        if matches:
            return matches.group(1)
        else:
            print("invalid url")



import asyncio
from datetime import datetime
import json
import math
import time
import requests
import spotipy
from shrillecho.spotify.client import SpotifyClient
from shrillecho.spotify.task_scheduler import TaskScheduler
from shrillecho.track.spotify_track import SpotifyTrack
from shrillecho.types.album_types import Album, ArtistAlbums, SeveralAlbums, SimplifiedAlbum, SimplifiedTrack
from shrillecho.types.artist_types import Artist, FollowerCount, SeveralArtists
from shrillecho.types.base_types import Followers
from shrillecho.types.track_types import SeveralTracks, Track
from shrillecho.user.user import SpotifyUserUtil
from shrillecho.utility.cache import cached
from shrillecho.utility.general_utility import get_id, is_earlier, sp_fetch
from typing import List



class SpotifyArtistUtil:

    @staticmethod
    def determine_main_artist(track: Track) -> Artist:
        """ The main artist is the artist who actually made the track, 
            so if its a remix , they are the creators of the remix but may not be listed
            as the first artist. The main artist does not include the feats just the core artist, 
            may extend to be feats..

            Args:
                track (Track): the track to determine the main artist of
        
            Returns:
                List[Artist]: list of artists in the playlist
        """

        main_artist = None

        if 'Remix' in track.name:
            for artist in track.artists:
                if artist.name in track.name:
                    main_artist = artist
            if not main_artist:
                main_artist = track.artists[0]
        else:
            main_artist = track.artists[0]

        return main_artist


    @staticmethod
    @cached(cache_key_prefix="followers", class_type=FollowerCount, expiry=3600)
    async def followers(sp: SpotifyClient, artist: str) -> int:
        """ the number of followers (ML form)

            Args:
                track (Artist): The artist to get the followers of (ML form)
        
            Returns:
                int: followers (ML form)
        """
        artist_id = get_id("artist", artist)
        data = {"artist": artist_id}
        followers_call = requests.post("http://localhost:8002/ml", json=data)
        followers = followers_call.json()
        return FollowerCount.from_json(json.dumps(followers))

    @staticmethod
    async def get_artist_track_ids(sp: SpotifyClient, artist_id) -> List[str]:
        """ Given an artist id, return all the associated track ids of that artist

            Args:
               sp (SpotifyClient): client
               artist_id (str): artist id (base64)
        
            Returns:
                int: followers (ML form)
        """
      
        artist_albums: List[SimplifiedAlbum] = await sp.artist_albums(artist=artist_id)
        
        task_scheduluer = TaskScheduler()

        tasks = [
            (sp.album_tracks, (album.id,)) for album in artist_albums
        ]
    
        task_scheduluer.load_tasks(tasks)

        
        album_tracks: List[SimplifiedTrack] = await task_scheduluer.execute_tasks(batch_size=20)
     
   
        track_ids = [item.id for item in album_tracks]

        return track_ids

    @staticmethod
    async def get_artist_tracks(sp: SpotifyClient, artist: str, unique: bool = True) -> List[Track]:
        """ Given an artist return all there tracks, unique is optional.

            Args:
               sp (SpotifyClient): client
               artist (str): artist id / uri / link
               unique: whether to remove duplicate tracks that appeared from the same release.
        
            Returns:
                int: followers (ML form)
        """
        
        artist_id = get_id("artist", artist)
        
        track_ids = await SpotifyArtistUtil.get_artist_track_ids(sp, artist_id)
     
        tracks: List[Track] = await SpotifyTrack.fetch_several_tracks(sp, track_ids)

        if unique:
            return set(tracks)
        
        return tracks
       
    
    @staticmethod
    async def get_artist_albums(sp: SpotifyClient, artist_id: str, simple=False) -> List[Album] | List[SimplifiedAlbum]:
        
        """ Given an artist id, return all albums """
        
        artist_albums: List[SimplifiedAlbum] = await sp.artist_albums(artist=artist_id)

        if simple:
            return artist_albums

        album_ids = [album.id for album in artist_albums]
        
        return SpotifyTrack.fetch_several_albums(sp, album_ids=album_ids)
    
    @staticmethod
    async def get_artist_new_releases(sp: SpotifyClient, artist_id: str, earliest_date: str ='2024-03-01' ) -> List[SimplifiedAlbum]:
       
        albums: List[SimplifiedAlbum] = []
        artist_albums: ArtistAlbums = await sp.artist_albums_single(artist=artist_id, offset=0)
        albums.extend(artist_albums.items)
        next = artist_albums.next
        
        while next != None:
           

            offset: int = artist_albums.offset + 50
            artist_albums: ArtistAlbums = await sp.artist_albums_single(artist=artist_id, offset=offset)

            if len(artist_albums.items) == 0:
                break

            last_album = artist_albums.items[-1]
            next = artist_albums.next

            if is_earlier(last_album.release_date, earliest_date):
                break

            albums.extend(artist_albums.items)

         
        return albums
    
    @staticmethod
    async def get_followed_new_releases(sp: SpotifyClient) -> List[SimplifiedAlbum]:
       
        followed_artists: List[Artist] = await SpotifyUserUtil.get_followed_artists(sp)
   

        ids = [artist.id for artist in followed_artists]

        task_scheduluer = TaskScheduler()

        tasks = [
            (SpotifyArtistUtil.get_artist_new_releases, (sp, artist_id,)) for artist_id in ids
        ]
    
        task_scheduluer.load_tasks(tasks)
   
        return await task_scheduluer.execute_tasks(batch_size=50)
     

    @staticmethod
    async def most_obscure_artists(sp: SpotifyClient, artist: str, artists: List[Artist], depth: int = 0) -> List[Artist]:
        
        if depth == 10:
            return artists

        artist_id = get_id("artist", artist)

        related_artists: List[Artist] = await sp.artist_related(artist=artist_id)
        
        most_obscure = math.inf
        most_obscure_artist = None

        for artist in related_artists:
            followers: FollowerCount = await SpotifyArtistUtil.followers(sp, artist=artist.id)
            follower_count = followers.followers

            if follower_count < most_obscure and artist not in artists:
                most_obscure = follower_count 
                most_obscure_artist = artist

        artists.append(most_obscure_artist)
        depth += 1

        return await SpotifyArtistUtil.most_obscure_artists(sp, most_obscure_artist.id, artists, depth)


    @staticmethod
    async def fetch_several_artists(sp: SpotifyClient, artist_ids: List[str]) -> List[Artist]:
        """ Fetch multiple artists

            Args:
                sp (SpotifyClient): client
                artist_ids (List[str]): list of ids to get the artist for
            
            Returns:
                List[Artist]: list of each artist
        """
        artists: List[Artist] = []
        for i in range(0, len(artist_ids), 50):
            chunk = artist_ids[i:i + 50]
            sev_artists: SeveralArtists = await sp.artists(chunk)
            artists.extend(sev_artists.artists)
        return artists

    @staticmethod
    async def get_playlist_artists(sp: SpotifyClient, playlist: str) -> List[Artist]:
        """ Given a playlist, returns all the artists on the playlist. Includes all artists of track and does not
            contain duplicate artists

            Args:
                sp (SpotifyClient): Spotify Client
                playlist (str): Playlist ID / URI / Link

            Returns:
                List[Artist]: list of artists in the playlist
        """

        playlist_id = get_id("playlist", playlist)

        playlist_tracks: List[Track] = await sp.playlist_tracks(playlist_id=playlist_id)

        artist_ids = set()

        for track in playlist_tracks:
            if track.artists and track.external_ids.isrc:
                for artist in track.artists:
                    artist_ids.add(artist.id)

        return SpotifyArtistUtil.fetch_several_artists(sp, list(artist_ids))
        


 

    




import json
from typing import List
import requests
from shrillecho.spotify.client import SpotifyClient
from shrillecho.spotify.task_scheduler import TaskScheduler
from shrillecho.track.spotify_track import SpotifyTrack
from shrillecho.types.album_types import SimplifiedAlbum, SimplifiedTrack
from shrillecho.types.artist_types import Artist, FollowerCount
from shrillecho.types.track_types import Track
from shrillecho.utility.cache import cached


class ArtistController:


    def __init__(self, sp: SpotifyClient, artist: Artist):
        self._artist = artist
        self._sp = sp


    @staticmethod
    async def create(sp: SpotifyClient, artist_id: str = None, artist: Artist = None):

        if not artist:
            artist: Artist = await sp.artist(artist=artist_id)

        return ArtistController(sp=sp, artist=artist)
    
    
    @cached(cache_key_prefix="followers", class_type=FollowerCount, expiry=3600)
    async def followers(self) -> int:
        """ the number of followers (ML form)

            Args:
                track (Artist): The artist to get the followers of (ML form)
        
            Returns:
                int: followers (ML form)
        """
        data = {"artist": self.artist.id}
        followers_call = requests.post("http://localhost:8002/ml", json=data)
        followers = followers_call.json()
        return FollowerCount.from_json(json.dumps(followers))
    
    async def get_artist_track_ids(self) -> List[str]:
        """ Return all tracks in the form of there ids

            Returns:
                List[str]: all artist tracks in for of ids
        """
      
        artist_albums: List[SimplifiedAlbum] = await self._sp.artist_albums(artist=self._artist)
        
        task_scheduluer = TaskScheduler()

        tasks = [
            (self._sp.album_tracks, (album.id,)) for album in artist_albums
        ]
    
        task_scheduluer.load_tasks(tasks)

        
        album_tracks: List[SimplifiedTrack] = await task_scheduluer.execute_tasks(batch_size=5)
     
   
        track_ids = [item.id for item in album_tracks]

        return track_ids

    async def get_artist_tracks(self, unique: bool = True) -> List[Track]:
        """ Return all tracks from an artist

            Args:
               unique: whether to remove duplicate tracks that appeared from the same release.
        
            Returns:
                int: followers (ML form)
        """
    
        tracks: List[Track] = await SpotifyTrack.fetch_several_tracks(sp=self._sp, track_ids= await self.get_artist_track_ids())

        if unique:
            return set(tracks)
        
        return tracks


from dataclasses import dataclass
from typing import Optional
# from dataclasses_json import dataclass_json
from fastclasses_json import dataclass_json


@dataclass_json
@dataclass
class ExternalUrls:
    spotify: Optional[str] = None
    href: Optional[str] = None



@dataclass_json
@dataclass
class ExternalIds:
    isrc: Optional[str] = None
    ean: Optional[str] = None
    upc: Optional[str] = None
 
   
  

@dataclass_json
@dataclass
class Followers:
    total: int


@dataclass_json
@dataclass
class Policies:
    opt_in_trial_premium_only_market: bool

@dataclass_json
@dataclass
class Copyright:
    text: str
    type: str

@dataclass_json
@dataclass
class ExplicitContent:
    filter_enabled: bool
    filter_locked: bool

@dataclass_json
@dataclass
class Image:
    url: str
    height: int
    width: int
   

@dataclass_json
@dataclass
class Restrictions:
    reason: Optional[str]
  



import os

from selenium import webdriver
from selenium.webdriver.firefox.service import Service
from selenium.webdriver.firefox.webdriver import WebDriver
from selenium.webdriver.support.ui import WebDriverWait
from shrillecho.utility.archive_maybe_delete import general as sh
import re


class Scraper:
    def __init__(self):
        self.__driver = self.setup_driver()

    def get_driver(self):
        return self.__driver

    def get(self, url):
        return self.__driver.get(url)
    

    def get_biscuits(self):
        return self.__driver.get_cookies()
    
    def setup_driver(self) -> WebDriver:
    
        firefox_options = webdriver.FirefoxOptions()
        # firefox_options.add_argument('--headless')

        service = Service()
        return webdriver.Firefox(service=service, options=firefox_options)

    def wait_on_load(self, driver, pattern):
        page_source = driver.page_source
        matches = set(re.findall(pattern, page_source))
        return len(matches) > 1
    
    def fetch_source(self, driver, uri, pattern):
        driver.get(f'{uri}')
        WebDriverWait(driver, 10).until(lambda driver=driver: self.wait_on_load(driver, pattern))
        return driver.page_source

    def get_followers(self, seed_user: str):
        base_uri = 'https://open.spotify.com/user'
        uri = f'{base_uri}/{seed_user}/followers'
        pattern = r'/user/(\w+)'
        source = self.fetch_source(self.__driver, uri, pattern)

        return set(re.findall(pattern, source))
    
    def add_cookies(self, sp_dc, sp_key):
        cookies = [
            {
                "name": "sp_dc",
                "value": sp_dc,
                "domain": "open.spotify.com"
            },
            {
                "name": "sp_key",
                "value": sp_key,
                "domain": "open.spotify.com"
            },
        ]

        for cookie in cookies:
            self.__driver.add_cookie(cookie)
        self.__driver.refresh()


